//-----------------------------------------------------------------------------
// SoC Labs Wrapper Files
// - Accelerator Packet Deconstruction from AHB Read Transactions
// A joint work commissioned on behalf of SoC Labs; under Arm Academic Access license.
//
// Contributors
//
// David Mapstone (d.a.mapstone@soton.ac.uk)
//
// Copyright 2023; SoC Labs (www.soclabs.org)
//-----------------------------------------------------------------------------
module  wrapper_packet_deconstruct #(
  parameter   ADDRWIDTH=11,
  parameter   PACKETWIDTH=256
)(
  input  logic                  hclk,       // clock
  input  logic                  hresetn,    // reset

   //Register interface
  input  logic [ADDRWIDTH-1:0]   addr,
  input  logic                   read_en,
  input  logic                   write_en,
  input  logic [3:0]             byte_strobe,
  input  logic [31:0]            wdata,
  output logic [31:0]            rdata,
  output logic                   wready,
  output logic                   rready,

  // Valid/Ready interface
  input  logic [PACKETWIDTH-1:0] packet_data,
  input  logic                   packet_data_last,
  input  logic                   packet_data_valid,
  output logic                   packet_data_ready
);

// Create Deconstruction Buffer
// logic [PACKETWIDTH-1:0] deconst_buffer;
logic [(PACKETWIDTH/32)-1:0][31:0] deconst_buf;

// Create Array to Flag which buffers have been read
logic [(PACKETWIDTH/32)-1:0] deconst_buf_flag;

// Curent Buffer Flag
logic [(PACKETWIDTH/32)-1:0] cur_deconst_buf_flag;
assign cur_deconst_buf_flag = 1'b1 << buf_word_sel;

// Check All Flags are High
logic deconst_buf_flag_reduced;
assign deconst_buf_flag_reduced = &(deconst_buf_flag | (cur_deconst_buf_flag));

// Select which word in buffer to read 
logic [$clog2(PACKETWIDTH/32)-1:0] buf_word_sel;
assign buf_word_sel = addr[($clog2(PACKETWIDTH/32)-1)+2:2];

logic deconst_buf_valid;
// Dump data on one of two conditions
// - An address ends [5:0] in 0x3C i.e. [5:2] == 0xF
// - Address Moved to different 512 bit word
// Write Condition
always_ff @(posedge hclk or negedge hresetn) begin
    if (~hresetn) begin
        // Reset Construction Buffer
        deconst_buf   <= {PACKETWIDTH{1'b0}};
        // Reset Values
        packet_data_ready  <= 1'b0;
        deconst_buf_valid  <= 1'b0;
        deconst_buf_flag   <= {(PACKETWIDTH/32){1'b0}};
    end else begin
        // If ready is low and theres no valid data in buffer, asser ready
        if (!packet_data_ready && !deconst_buf_valid) begin
            packet_data_ready  <= 1'b1;
        end
        // Read Packet into Deconstruction Buffer
        if (packet_data_valid && packet_data_ready) begin
            packet_data_ready  <= 1'b0;
            deconst_buf        <= packet_data;
            deconst_buf_valid  <= 1'b1;
            deconst_buf_flag   <= {(PACKETWIDTH/32){1'b0}};
        end
        if (read_en) begin
            // Register which words in the Deconstruction buffer have been read
            // Check if All Words have been Read
            if (deconst_buf_flag_reduced && !(packet_data_valid && packet_data_ready)) begin
                // Set Ready High To Get more Data into Buffer
                deconst_buf_valid  <= 1'b0;
                packet_data_ready  <= 1'b1;
            end else begin
                deconst_buf_flag   <= deconst_buf_flag | cur_deconst_buf_flag;
            end
        end
    end
end

// Read Condition
always_comb begin
    if (read_en) begin
        // Read appropriate 32 bits from buffer - wrapping behaviour
        rdata = deconst_buf[buf_word_sel];
    end else begin
        rdata = 32'd0;
    end
end

// Register Ready Control
always_comb begin
    // Not Ready Out when waiting for Valid Data on Input
    rready = ~packet_data_ready;
    // Write Ready always high but doesn't do anywthing
    wready = 1'b1;
end

endmodule